We've conflated the default parameter problem with the concrete interface 
problem and are trying to solve them simultaneously.

## Defaulted parameters

Why not use Java overloading for handling defaulted parameters.

* Class methods:
    * Each defaulted parameter corresponds to a 
     `final` overloaded method on the same class, with the same visibility as 
     the no-defaults method. 
    * The overloaded methods have the same type parameters and 
      bounds as the no-defaults method. 
    * The overloaded methods have the `@Ignore` annotation.

* Class initializers:
    * Each defaulted parameter is given by a `final` overloaded constructor on 
    the same class, with the same visibility as the no-defaults constructor.

* Top-level methods:
    * Each defaulted parameter is given by a `final` overloaded method on the 
      same class, with the same visibility as the no-defaults method. 
    * The overloaded methods have the same type parameters and bounds as the 
      no-defaults method. 
    * The overloaded methods have the `@Ignore` annotation.

* Interface methods:
    * Each defaulted parameter corresponds to an overloaded method on the same 
      interface, with the same visibility as the no-defaults method. 
    * The default value expressions would be methods in the companion class 
      (as for concrete interface members).
    * The overloaded methods  have the same type parameters and bounds as the 
      no-defaults method. 
    * The overloaded methods have the `@Ignore` annotation.

## Concrete interface members

In Java any type declared within an interface is implicitly `static`, but in 
Ceylon it is not. This causes major headaches. Simiarly any inner interface is 
also implicitly static. This causes migraines.

* Class members of interfaces: 
    * Are generated at the place where the Ceylon interface is declared.
    * Have all type parameters (including bounds) copied from the outer 
      type(s).
    * Have generated constructors (plural, they could be overloaded due to 
      defaulted parameters) that take the outer instance as their first 
      parameter.
    * Have an `$outer()` method generated which returns the outer instance.
* Interface members of interfaces: **TODO**
* Method members of interfaces have implementation added to the companion 
  class.
* Object members of interfaces are treated as class members of interfaces
* Attribute members of interfaces have implementation added to the companion 
  class.

The generated companion class (`$impl`):

* Is generated at the place where the Ceylon interface is declared.
* Have all type parameters (including bounds) copied from the outer type(s).
* Have generated constructor that takes the interface instance and 
  any outer instance as their first parameters.
* Have a `$this` field which holds the interface instance.
* Have an `$outer()` method generated which returns the outer instance. 

The generated interfaces:

* Are *always* top level (because inner interfaces are implicitly 
  `static` and therefore you can't have interfaces inner to an inner type).

### On the `static`-ness of `$impl`s

It would be nice if we could make `$impl` classes `static`, and pass 
the outer reference as a secret first argument to each of the methods. 
This doesn't work in general because in Java you can't have 
`static` second level types, so they would have to be top-level classes 
but then you loose access to `private` members of outer types (see the second 
example below).

Since they're not `static` an `$impl` would need to be instantiated whenever 
a direct implementation of the interface was instantiated:

    class Implementor implements I {
        private final I$impl $I$impl;
        Implementor() {
            $I$impl = new I$impl(this);
        }
        // ...
    }

### On the transformation of self references

The simplest thing to do is to always generate an `$outer()` method (on Java 
classes and interfaces corresponding to nested Ceylon declarations). For
normal classes the method can use a Java-style `Outer.this`, and for 
`$impl` classes we return this `$outer` field.
 
Maybe one day it will be possible to inline these and avoid generating and 
calling methods entirely.

We will also need to do something special about `this`, 
otherwise a `this` within a concrete interface method will 
accidentally refer to the `$impl` and not the actual instance of the 
implementation. I don't think we need to encapsulate `$this` in a method 
though.

We do not need to worry about `super` because it's not permitted in an 
interface body.

## On local declarations

Local declarations are potentially a problem because in Ceylon you 
could capture a reference from the containing method or initializer
for use in a concrete interface method or defaulted parameter.

However this should not be a problem because the `$impl` classes 
(in the case of interfaces) or the local classes (in the case of classes) is 
located where the Ceylon type declaration is made, so will be able to capture 
the same variables.

### Examples

Let us now consider all 8 combinations of classes and/or interfaces nested 
three deep...

# Putting it all together

* Transforming a top level class: Trivial
* Transforming a top level interface: 
    * Generate a `$impl` immediately after the transformed interface
    * Give it the same type parameters as the interface
    * Give it a constructor which takes the interface instance; assign it to 
      the `$this` field
    * When transforming `this` expressions within the interface, generate 
      `$this`
* Transforming a top-level method (function):
    * Generate a wrapper class (as now)
    * Generate overloaded methods for defaulted parameters (with the same type 
      parameters as the no-defaults method)
* When transforming a class inner to a class:
    * Trivial
* When transforming an interface inner to a class:
    * Generate a `$impl` where the interface was declared.
    * Give it the same type parameters as the interface
    * Give it a constructor which takes the interface instance; assign it to 
      the `$this` field
    * Generate a `$outer()` using `Outer.this` syntax
    * When transforming `this` expressions within the interface, generate 
      `$this`
    * When transforming `outer` expressions within the interface, generate 
      `$outer()`
    * Generate the interface immediately after the class.
    * Prepend all the type parameters and bounds from the containing class 
      (and its containing type(s))
    * Then give it the same type parameters as the declared interface.
    
* When transforming a class inner to an interface:
* When transforming an interface inner to an interface:
* When transforming a local interface:
    * Generate the interface top level (add some name mangling)
    * Generate 
* When transforming a local class:

# Refactoring plan

1. DONE Start generating overloaded methods for defaulted parameters.
1. DONE Change positional callsites for defaulted parameters to use the 
   overloaded methods.
1. DONE? For class methods and functions generate the default expression methods
   on the class itself. And fix call sites.
   NONE-`this` INVOCATIONS NEED TESTING
   
1. DONE Stop generating `$impl` for classes and functions etc.
1. DONE Making interface `$impl` non-static and move where we declare it
1. DONE For interface `$impl`s generate constructor which takes a `$this` and `$outer`
1. DONE Generate `$outer()` on inner interfaces
1. DONE For class initializers generate a non-`static` sibling `$impl`, which 
   doesn't take a `$this` in its constructor. 
   And has the same type parameters. 
   And fix call sites to generate a Let which calls super
1. DONE When implementing an interface containing methods with default params, 
   generate: 
    * the overload methods,
    * default value method implementations and
    * `$outer()`  method implementation.
1. DONE Instantiate interface `$impl`s when needed (i.e. when instantiating a 
   direct implementor). ??Do this even if the interface currently doesn't 
   strictly need a companion class for bin compat purposes??
1. DONE Refactor names
1. DONE Move generated interfaces to top level (with name mangling) (remember, they 
   can't be left where they are because they're implicitly static and static 
   is disallowed in second level decls).
1. DONE When generating interfaces copy down type parameters of qualifying types.
1. DONE Tests using parameter bounds 
1. Use capture infomation to avoid generating a `$impl` delegate method if it's 
   not required. (Not sure)
1. DONE Fix local declarations
1. Rename top level types in tests for Tako
1. Fix model loader
1. Tests with sequenced parameters.
1. Tests with intersection/union types
1. Test Non-outer type parameters
1. More tests of interface instantiations (esp inner interface instantiation)
1. Test formal member classes.
1. Add in at()s 
1. Make all builders double-build() safe, and methodbuilder.noBody()
1. Review Added TODOs and XXXs
1. Write more tests for nested interfaces and concrete interface members
1. Mostly DONE Tidy up the patch set
1. DONE Fix all the real tests.
1. Merge into origin/master
1. Optimisation: Use `static` `$impl` classes for top and first level declarations.
1. Optimization: If a class implements all the methods on an interface we don't 
   need to instantiate a $impl

# Local declarations

* We cannot ascribe a type to `$outer()` for inner interfaces of local types
  because we've hoisted the actual interface type to top level (so it's not 
  visible at the point of declaration. We can't not hoist it because of the 
  staticness of interfaces problem. So maybe we should just cast it

# Optimizations and BC

Tako asked last night about what optimisations we can make to avoid unnecessary 
$impl allocations. My reply at the time was that I thought it was possible for 
top-level types (with no inner or local classes). That has a problem though in 
that it breaks BC if you add an inner (because the calling convention changes):

interface I {
    shared formal void m(Integer i = 1);
    // Calling convention would use a static m$i(I $this)
}

Changed to:

interface I {
    shared formal void m(Integer i = 1);
    // Calling convention would now use a non-static m$i()
    // (consider I nested within something else to see why it cannot be static)
    class C() {
        void m2() {// Closure over m may or may not be required
            m();
        }
    }
}

However, as well as the interface $impl, we also instantiate constructor 
`$impls`, and I think these *can* be optimized safely for top level types.

class C(Integer i = 0){
}

Translating to:

class C {
    static Integer $init$i() {
        return 0;
    }
    C(Integer i) {
    }
    C() {
        this($init$i());
    }
}

Again because of the nested static problem this only applies to top level 
classes.

Now, for interface methods with defaulted parameters (and not just top level), 
it should be possible to push the default value methods to the top level 
class or `$impl` (with suitable name munging), and make them static, 
copying down type parameters and passing `$this` variables.

But I don't think that really works for concrete interface members, because of 
the capture issue. Maybe for non-shared inner interfaces we can do 
it if they're not captured.

Tako also thought it's too soon to think about BC. For BC between compiler 
realeases I  would agree. But for BC between module versions compiled with the 
same compiler, I think we have to worry about that now.

# Problems

The relocating of interfaces has some nasty consequences:

* Interface members are public, but because they're now top level they're 
   public to anyone. I don't think this is really a problem because 
   I don't think the JRE prevents you from access member interfaces if you know 
   the munged name.
* The name munging means the ceylon name is not the same as the java name
  so we might need a 'name' attribute on the @Ceylon annotation of the 
  model loader

# Encapsulations

* Generating the FQ and unqualified companion class name
* Generating JCExpression for defaultValue method name

# TODO

* There's a problem with intersection and unions involving type parameters 
  when we figure out the type parameterization of methods etc in
  AbstractTransformer.getTypeForParameter() 
  (Use of ThisType() in the InvocationBuilder)
* Think some more about bin compat for interface methods. 


# Interop

The problem is that Java (and for that matter `ceylon.language`) interfaces
don't come with `$impl` classes.

When compiling class `C` satisfying interface `I` we could check whether`I$impl` 
exists and generate code to instantiate it only if it does. But we must consider
what modifications to I would cause it to have a $impl -- are they **only** 
modifications which would otherwise be non-BC anyway?

* Adding a default to an existing param
* Adding a new defaulted param to an existing method (wouldn't be BC anyway)
* Adding a new method with defaulted parameters (OK, wouldn't be BC anyway)
* Adding a new method with defaulted parameters (OK, wouldn't be BC anyway)
* Adding an inner type (in this case we need the `$impl` for capture purposes)
  but the presence of that inner type might not be apparent to interface 
  implementors
